# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import importlib
import logging
import os
import re
from contextlib import contextmanager
from pathlib import Path
from time import monotonic
from typing import Callable, Generator, List, Optional, Sequence, Tuple

import libcst as cst

Timing = Tuple[str, float]

LOG = logging.getLogger(__name__)
LIBCST_PARSER_TYPE = "LIBCST_PARSER_TYPE"
INLINE_COMMENT_RE = re.compile(r"#+[^#]*")
TIMINGS: List[Timing] = []


@contextmanager
def timed(msg: str) -> Generator[None, None, None]:
    """
    Records the monotonic duration of the contained context, with a given description.

    Timings are stored for later use/printing with `print_timings()`.
    """
    before = monotonic()
    yield
    after = monotonic()
    TIMINGS.append((msg, after - before))


def get_timings() -> Sequence[Tuple[str, float]]:
    try:
        return list(TIMINGS)
    finally:
        TIMINGS.clear()


def print_timings(
    fn: Callable[[str], None] = print, *, timings: Sequence[Timing]
) -> None:
    """
    Print all stored timing values in microseconds.
    """
    for msg, duration in timings:
        fn(f"{msg + ':':50} {int(duration*1000000):7} µs")


def enable_libcst_native() -> bool:
    """
    Attempt to enable LibCST's faster, native PEG parser if available.

    This enables parsing files with 3.10 match statements, or other new syntax features
    dependent on having a PEG style parser. This is currently not the default parser
    in LibCST, so this function must be called before parsing files with 3.10+ syntax,
    or otherwise set the environment variable ``LIBCST_PARSER_TYPE="native"``.

    Returns ``True`` if the native parser was found and enabled, or ``False`` otherwise.
    """

    try:
        LOG.debug("checking for libcst native parser")
        # use importlib so that we can test/mock import errors
        module = importlib.import_module("libcst.native")

        if hasattr(module, "parse_module"):
            os.environ[LIBCST_PARSER_TYPE] = "native"
            LOG.debug("libcst native parser enabled")
            return True

    except ImportError:
        LOG.debug("libcst native module not available")

    return False


def try_parse(path: Path, data: Optional[bytes] = None) -> cst.Module:
    """
    Attempts to parse the file with all syntax versions known by LibCST.

    If parsing fails on all supported grammar versions, then raises the parser error
    from the first/newest version attempted.
    """
    if data is None:
        data = path.read_bytes()

    with timed(f"parsing {path}"):
        parse_error: Optional[cst.ParserSyntaxError] = None

        for version in cst.KNOWN_PYTHON_VERSION_STRINGS[::-1]:
            try:
                mod = cst.parse_module(
                    data, cst.PartialParserConfig(python_version=version)
                )
                return mod
            except cst.ParserSyntaxError as e:
                # keep the first error we see in case parsing fails on all versions
                if parse_error is None:
                    parse_error = e

        # not caring about existing traceback here because it's not useful for parse
        # errors, and usort_path is already going to wrap it in a custom class
        raise parse_error or Exception("unknown parse failure")


def parse_import(code: str) -> cst.SimpleStatementLine:
    """
    Parse a single import statement. For testing and debugging purposes only.
    """
    node = cst.parse_statement(code)
    if not isinstance(node, cst.SimpleStatementLine):
        raise ValueError("not a statement")
    if not isinstance(node.body[0], (cst.Import, cst.ImportFrom)):
        raise ValueError("not an import statement")
    return node


def split_inline_comment(text: str) -> Sequence[str]:
    return [part.rstrip() for part in INLINE_COMMENT_RE.findall(text)]


def split_relative(name: str) -> Tuple[str, int]:
    ndots = len(name) - len(name.lstrip("."))
    return name[ndots:], ndots


def stem_join(stem: Optional[str], name: str) -> str:
    if stem is None:
        return name
    elif stem.endswith("."):
        return stem + name
    else:
        return f"{stem}.{name}"


def top_level_name(name: str) -> str:
    return name.split(".", 1)[0]


def with_dots(x: cst.CSTNode) -> str:
    """
    Helper to make it easier to use an Attribute or Name.
    """
    if isinstance(x, cst.Attribute):
        return ".".join([with_dots(x.value), with_dots(x.attr)])
    elif isinstance(x, cst.Name):
        return x.value
    else:
        raise TypeError(f"Can't with_dots on {type(x)}")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from pathlib import Path

import libcst as cst

from ..config import Config
from ..sorting import ImportSorter


class SplitTest(unittest.TestCase):
    def test_split_block(self) -> None:
        mod = cst.parse_module(
            b"""\
from a import x
from b import y
from b import y
from c import x
"""
        )
        x = ImportSorter(module=mod, path=Path(), config=Config())
        blocks = x.sortable_blocks(mod.children)  # type: ignore
        self.assertEqual(2, len(blocks))
        self.assertEqual({"x": "a.x"}, blocks[0].imported_names)
        self.assertEqual({"y": "b.y", "x": "c.x"}, blocks[1].imported_names)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Sequence, Union

import libcst as cst

from .config import Config
from .types import (
    COMMENT_INDENT,
    ImportComments,
    ImportItemComments,
    SortableImport,
    SortableImportItem,
)
from .util import split_inline_comment, split_relative, with_dots


def render_node(node: cst.CSTNode, module: Optional[cst.Module] = None) -> str:
    if module is None:
        module = cst.Module(body=[])
    code = module.code_for_node(node.with_changes(leading_lines=()))
    return code


def name_to_node(name: str) -> Union[cst.Name, cst.Attribute]:
    if "." not in name:
        return cst.Name(name)

    base, name = name.rsplit(".", 1)
    return cst.Attribute(value=name_to_node(base), attr=cst.Name(name))


def import_comments_from_node(node: cst.SimpleStatementLine) -> ImportComments:
    comments = ImportComments()

    assert len(node.body) == 1, "lines with multiple statements are unsupported"
    assert isinstance(node.body[0], (cst.Import, cst.ImportFrom))
    imp: Union[cst.Import, cst.ImportFrom] = node.body[0]

    # # THIS PART
    # import foo
    #
    # # THIS PART
    # from foo import bar
    for line in node.leading_lines:
        if line.comment:
            comments.before.append(line.comment.value)
        else:
            comments.before.append("")

    if isinstance(imp, cst.ImportFrom):
        if imp.lpar:
            if isinstance(imp.lpar.whitespace_after, cst.ParenthesizedWhitespace):
                ws = imp.lpar.whitespace_after

                # from foo import (  # THIS PART
                #     bar,
                # )
                if ws.first_line.comment:
                    comments.first_inline.extend(
                        split_inline_comment(ws.first_line.comment.value)
                    )

                # from foo import (
                #     # THIS PART
                #     bar,
                # )
                comments.initial.extend(
                    line.comment.value for line in ws.empty_lines if line.comment
                )

            assert imp.rpar is not None
            if isinstance(imp.rpar.whitespace_before, cst.ParenthesizedWhitespace):
                # from foo import (
                #     bar
                #     # THIS PART
                # )
                comments.final.extend(
                    line.comment.value
                    for line in imp.rpar.whitespace_before.empty_lines
                    if line.comment
                )

                # from foo import (
                #     bar,
                #     baz  # THIS PART (NO COMMA!)
                # )
                if imp.rpar.whitespace_before.first_line.comment:
                    comments.inline.extend(
                        split_inline_comment(
                            imp.rpar.whitespace_before.first_line.comment.value
                        )
                    )

            # from foo import (
            #     bar,
            # )  # THIS PART
            if node.trailing_whitespace and node.trailing_whitespace.comment:
                comments.last_inline.extend(
                    split_inline_comment(node.trailing_whitespace.comment.value)
                )

        # from foo import bar  # THIS PART
        elif node.trailing_whitespace and node.trailing_whitespace.comment:
            comments.first_inline.extend(
                split_inline_comment(node.trailing_whitespace.comment.value)
            )

    # import foo  # THIS PART
    elif isinstance(imp, cst.Import):
        if node.trailing_whitespace and node.trailing_whitespace.comment:
            comments.first_inline.extend(
                split_inline_comment(node.trailing_whitespace.comment.value)
            )

    else:
        raise TypeError

    return comments


def item_from_node(
    node: cst.ImportAlias, stem: Optional[str] = None, before: Sequence[str] = ()
) -> SortableImportItem:
    name = with_dots(node.name)
    asname = with_dots(node.asname.name) if node.asname else ""
    comments = ImportItemComments()
    comments.before.extend(before)
    comma = False

    if isinstance(node.comma, cst.Comma):
        comma = True
        if (
            isinstance(node.comma.whitespace_before, cst.ParenthesizedWhitespace)
            and node.comma.whitespace_before.first_line.comment
        ):
            # from foo import (
            #     bar  # THIS PART
            #     ,
            # )
            comments.inline.extend(
                split_inline_comment(
                    node.comma.whitespace_before.first_line.comment.value
                )
            )

        if isinstance(node.comma.whitespace_after, cst.ParenthesizedWhitespace):
            ws = cst.ensure_type(
                node.comma.whitespace_after, cst.ParenthesizedWhitespace
            )
            if ws.first_line.comment:
                # from foo import (
                #     baz,  # THIS PART
                # )
                comments.inline.extend(
                    split_inline_comment(ws.first_line.comment.value)
                )

            # from a import (
            #   b,
            #   # THIS PART
            #   c,  # (but only if it's not the last item)
            # )
            comments.following.extend(
                line.comment.value for line in ws.empty_lines if line.comment
            )

    return SortableImportItem(
        name=name, asname=asname, comments=comments, stem=stem, comma=comma
    )


def import_from_node(node: cst.SimpleStatementLine, config: Config) -> SortableImport:
    # TODO: This duplicates (differently) what's in the LibCST import metadata visitor.

    stem: Optional[str] = None
    items: List[SortableImportItem] = []
    comments = import_comments_from_node(node)

    # There are 4 basic types of import
    # Additionally some forms z can have leading dots for relative
    # imports, and there can be multiple on the right-hand side.
    #
    if isinstance(node.body[0], cst.Import):
        # import z
        # import z as y
        items = [item_from_node(name) for name in node.body[0].names]

    elif isinstance(node.body[0], cst.ImportFrom):
        # from z import x
        # from z import x as y

        # This is treated as a barrier and should never get this far.
        assert not isinstance(node.body[0].names, cst.ImportStar)

        stem = with_dots(node.body[0].module) if node.body[0].module else ""

        if node.body[0].relative:
            stem = "." * len(node.body[0].relative) + stem

        for name in node.body[0].names:
            items.append(item_from_node(name, stem, comments.initial))
            comments.initial = []

    else:
        raise TypeError

    # assume that "following" comments are actually meant for an item after that
    prev: Optional[SortableImportItem] = None
    for item in items:
        if prev is not None and prev.comments.following:
            item.comments.before.extend(prev.comments.following)
            prev.comments.following.clear()
        prev = item
    # inline comments after the last import (without a comma)
    if prev is not None and comments.inline:
        prev.comments.inline.extend(comments.inline)
        comments.inline.clear()
    # following comments on the last item should maybe be on the import itself
    if prev is not None and prev.comments.following:
        comments.final.extend(prev.comments.following)
        prev.comments.following.clear()

    imp = SortableImport(
        stem=stem,
        items=items,
        comments=comments,
        config=config,
        node=node,
        indent="",
    )
    return imp


def import_to_node(
    imp: SortableImport, module: cst.Module, indent: str, config: Config
) -> cst.BaseStatement:
    if config.magic_commas and imp.stem and imp.trailing_comma:
        return import_to_node_multi(imp, module)
    node = import_to_node_single(imp, module)
    content = indent + render_node(node, module).rstrip()
    # basic imports can't be reflowed, so only deal with from-imports
    if imp.stem and len(content) > config.line_length:
        node = import_to_node_multi(imp, module)
    return node


def import_to_node_single(imp: SortableImport, module: cst.Module) -> cst.BaseStatement:
    leading_lines = [
        (
            cst.EmptyLine(indent=True, comment=cst.Comment(line))
            if line.startswith("#")
            else cst.EmptyLine(indent=False)
        )
        for line in imp.comments.before
    ]

    trailing_whitespace = cst.TrailingWhitespace()
    trailing_comments = list(imp.comments.first_inline)

    names: List[cst.ImportAlias] = []
    for item in imp.items:
        name = name_to_node(item.name)
        asname = cst.AsName(name=cst.Name(item.asname)) if item.asname else None
        node = cst.ImportAlias(name=name, asname=asname)
        names.append(node)
        trailing_comments += item.comments.before
        trailing_comments += item.comments.inline
        trailing_comments += item.comments.following

    trailing_comments += imp.comments.inline
    trailing_comments += imp.comments.final
    trailing_comments += imp.comments.last_inline
    if trailing_comments:
        text = COMMENT_INDENT.join(trailing_comments)
        trailing_whitespace = cst.TrailingWhitespace(
            whitespace=cst.SimpleWhitespace(COMMENT_INDENT), comment=cst.Comment(text)
        )

    if imp.stem:
        stem, ndots = split_relative(imp.stem)
        if not stem:
            module_name = None
        else:
            module_name = name_to_node(stem)
        relative = (cst.Dot(),) * ndots

        line = cst.SimpleStatementLine(
            body=[cst.ImportFrom(module=module_name, names=names, relative=relative)],
            leading_lines=leading_lines,
            trailing_whitespace=trailing_whitespace,
        )

    else:
        line = cst.SimpleStatementLine(
            body=[cst.Import(names=names)],
            leading_lines=leading_lines,
            trailing_whitespace=trailing_whitespace,
        )

    return line


def import_to_node_multi(imp: SortableImport, module: cst.Module) -> cst.BaseStatement:
    body: List[cst.BaseSmallStatement] = []
    names: List[cst.ImportAlias] = []
    prev: Optional[cst.ImportAlias] = None
    following: List[str] = []
    lpar_lines: List[cst.EmptyLine] = []
    lpar_inline: cst.TrailingWhitespace = cst.TrailingWhitespace()

    item_count = len(imp.items)
    for idx, item in enumerate(imp.items):
        name = name_to_node(item.name)
        asname = cst.AsName(name=cst.Name(item.asname)) if item.asname else None

        # Leading comments actually have to be trailing comments on the previous node.
        # That means putting them on the lpar node for the first item
        if item.comments.before:
            lines = [
                cst.EmptyLine(
                    indent=True,
                    comment=cst.Comment(c),
                    whitespace=cst.SimpleWhitespace(module.default_indent),
                )
                for c in item.comments.before
            ]
            if prev is None:
                lpar_lines.extend(lines)
            else:
                prev.comma.whitespace_after.empty_lines.extend(lines)  # type: ignore

        # all items except the last needs whitespace to indent the *next* line/item
        indent = idx != (len(imp.items) - 1)

        first_line = cst.TrailingWhitespace()
        inline = COMMENT_INDENT.join(item.comments.inline)
        if inline:
            first_line = cst.TrailingWhitespace(
                whitespace=cst.SimpleWhitespace(COMMENT_INDENT),
                comment=cst.Comment(inline),
            )

        if idx == item_count - 1:
            following = (
                item.comments.following + imp.comments.inline + imp.comments.final
            )
        else:
            following = item.comments.following

        after = cst.ParenthesizedWhitespace(
            indent=True,
            first_line=first_line,
            empty_lines=[
                cst.EmptyLine(
                    indent=True,
                    comment=cst.Comment(c),
                    whitespace=cst.SimpleWhitespace(module.default_indent),
                )
                for c in following
            ],
            last_line=cst.SimpleWhitespace(module.default_indent if indent else ""),
        )

        node = cst.ImportAlias(
            name=name,
            asname=asname,
            comma=cst.Comma(whitespace_after=after),
        )
        names.append(node)
        prev = node

    # from foo import (
    #     bar
    # )
    if imp.stem:
        stem, ndots = split_relative(imp.stem)
        if not stem:
            module_name = None
        else:
            module_name = name_to_node(stem)
        relative = (cst.Dot(),) * ndots

        # inline comment following lparen
        if imp.comments.first_inline:
            inline = COMMENT_INDENT.join(imp.comments.first_inline)
            lpar_inline = cst.TrailingWhitespace(
                whitespace=cst.SimpleWhitespace(COMMENT_INDENT),
                comment=cst.Comment(inline),
            )

        body = [
            cst.ImportFrom(
                module=module_name,
                names=names,
                relative=relative,
                lpar=cst.LeftParen(
                    whitespace_after=cst.ParenthesizedWhitespace(
                        indent=True,
                        first_line=lpar_inline,
                        empty_lines=lpar_lines,
                        last_line=cst.SimpleWhitespace(module.default_indent),
                    ),
                ),
                rpar=cst.RightParen(),
            )
        ]

    # import foo
    else:
        raise ValueError("can't render basic imports on multiple lines")

    # comment lines above import
    leading_lines = [
        (
            cst.EmptyLine(indent=True, comment=cst.Comment(line))
            if line.startswith("#")
            else cst.EmptyLine(indent=False)
        )
        for line in imp.comments.before
    ]

    # inline comments following import/rparen
    if imp.comments.last_inline:
        inline = COMMENT_INDENT.join(imp.comments.last_inline)
        trailing = cst.TrailingWhitespace(
            whitespace=cst.SimpleWhitespace(COMMENT_INDENT), comment=cst.Comment(inline)
        )
    else:
        trailing = cst.TrailingWhitespace()

    return cst.SimpleStatementLine(
        body=body,
        leading_lines=leading_lines,
        trailing_whitespace=trailing,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys
from functools import partial
from pathlib import Path
from typing import Iterable, Optional, Set, Tuple, Union
from warnings import warn

from trailrunner import run, walk

from .config import Config
from .sorting import ImportSorter
from .types import Result
from .util import get_timings, timed, try_parse


__all__ = ["usort_bytes", "usort_string", "usort_path", "usort_stdin"]


def usort(data: bytes, config: Config, path: Optional[Path] = None) -> Result:
    """
    Given bytes for a module, this parses and sorts imports, and returns a Result.
    """
    if path is None:
        path = Path("<data>")

    try:
        module = try_parse(data=data, path=path)
        sorter = ImportSorter(module=module, path=path, config=config)
        new_mod = sorter.sort_module()

        return Result(
            path=path,
            content=data,
            output=new_mod.bytes,
            encoding=new_mod.encoding,
            timings=get_timings(),
            warnings=sorter.warnings,
        )

    except Exception as e:
        return Result(
            path=path,
            content=data,
            error=e,
            timings=get_timings(),
        )


def usort_bytes(
    data: bytes, config: Config, path: Optional[Path] = None
) -> Tuple[bytes, str]:
    """
    Returns (new_bytes, encoding_str) after sorting.

    DEPRECATED: use `usort()` directly instead.
    """
    warn("use usort() instead", DeprecationWarning)

    result = usort(data=data, config=config, path=path)
    if result.error:
        raise result.error

    assert result.encoding is not None
    return result.output, result.encoding


def usort_string(data: str, config: Config, path: Optional[Path] = None) -> str:
    r"""
    Whenever possible use usort_bytes instead.

    One does not just .read_text() Python source code.  That will use the system
    encoding, which if is not utf-8 would be in violation of pep 3120.

    There are two additional cases where this function does the wrong thing, but you
    won't notice on most modern file contents:

    - a string unrepresentable in utf-8, e.g. "\ud800" is a single high surrogate
    - a string with a valid pep 263 coding line, other than utf-8

    DEPRECATED: use `usort()` directly, and and encode/decode bytes as necessary.
    """
    warn("use usort() instead", DeprecationWarning, stacklevel=2)

    result = usort(data=data.encode(), config=config, path=path)
    if result.error:
        raise result.error

    return result.output.decode()


def usort_file(path: Path, *, write: bool = False) -> Result:
    """
    Format a single file and return a Result object.

    Ignores any configured :py:attr:`excludes` patterns.
    """

    try:
        config = Config.find(path.parent)
        data = path.read_bytes()
        result = usort(data, config, path)

        if result.output and write:
            path.write_bytes(result.output)

        return result

    except Exception as e:
        return Result(
            path=path,
            error=e,
            timings=get_timings(),
        )


def usort_path(
    paths: Union[Path, Iterable[Path]], write: bool = False
) -> Iterable[Result]:
    """
    For a given path, format it, or any python files in it, and yield :class:`Result` s.

    If given a directory, it will be searched, recursively, for any Python source files,
    excluding any files or directories that match the project root's ``.gitignore`` or
    any configured :py:attr:`excludes` patterns in the associated ``pyproject.toml``.
    """
    if isinstance(paths, Path):
        source_paths: Iterable[Path] = [paths]
    else:
        source_paths = paths

    with timed("total"):
        materialized_paths: Set[Path] = set()

        for path in source_paths:
            with timed(f"walking {path}"):
                config = Config.find(path)
                materialized_paths.update(walk(path, excludes=config.excludes))

        fn = partial(usort_file, write=write)
        if len(materialized_paths) == 1:
            # shave off multiprocessing overhead
            results = [fn(materialized_paths.pop())]
        else:
            results = [v for v in run(materialized_paths, fn).values()]
        return results


def usort_stdin() -> bool:
    """
    Read file contents from stdin, format it, and write the resulting file to stdout

    In case of error during sorting, no output will be written to stdout, and the
    exception will be written to stderr instead.

    Returns True if formatting succeeded, otherwise False
    """
    if sys.stdin.isatty():
        print("Warning: stdin is a tty", file=sys.stderr)

    try:
        config = Config.find()
        data = sys.stdin.read()
        result = usort_string(data, config, Path("<stdin>"))
        sys.stdout.write(result)
        return True

    except Exception as e:
        sys.stderr.write(repr(e))
        return False



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import unittest
from typing import Sequence
from unittest.mock import Mock, patch

import libcst as cst

from .. import util


class UtilTest(unittest.TestCase):
    def test_enable_libcst_native(self) -> None:
        with self.subTest("success"):
            os.environ.pop(util.LIBCST_PARSER_TYPE, None)

            module_mock = Mock(["parse_module"])
            import_mock = Mock(return_value=module_mock)
            with patch("usort.util.importlib.import_module", import_mock):
                result = util.enable_libcst_native()

                import_mock.assert_called_with("libcst.native")
                self.assertIn(util.LIBCST_PARSER_TYPE, os.environ)
                self.assertEqual(os.environ[util.LIBCST_PARSER_TYPE], "native")
                self.assertTrue(result)

        with self.subTest("import error"):
            os.environ.pop(util.LIBCST_PARSER_TYPE, None)

            import_mock = Mock(side_effect=ImportError())
            with patch("usort.util.importlib.import_module", import_mock):
                result = util.enable_libcst_native()

                import_mock.assert_called_with("libcst.native")
                self.assertNotIn(util.LIBCST_PARSER_TYPE, os.environ)
                self.assertFalse(result)

        with self.subTest("missing parse_module"):
            os.environ.pop(util.LIBCST_PARSER_TYPE, None)

            module_mock = Mock([])
            import_mock = Mock(return_value=module_mock)
            with patch("usort.util.importlib.import_module", import_mock):
                result = util.enable_libcst_native()

                import_mock.assert_called_with("libcst.native")
                self.assertNotIn(util.LIBCST_PARSER_TYPE, os.environ)
                self.assertFalse(result)

    def test_parse_import_simple(self) -> None:
        node = util.parse_import("import a")
        self.assertEqual(
            cst.ensure_type(
                cst.ensure_type(
                    cst.ensure_type(node, cst.SimpleStatementLine).body[0],
                    cst.Import,
                ).names[0],
                cst.ImportAlias,
            ).name.value,
            "a",
        )

    def test_parse_import_from(self) -> None:
        node = util.parse_import("from a import x")
        inner = cst.ensure_type(
            cst.ensure_type(node, cst.SimpleStatementLine).body[0],
            cst.ImportFrom,
        )
        self.assertEqual(cst.ensure_type(inner.module, cst.Name).value, "a")
        assert isinstance(inner.names, Sequence), f"{type(inner.names)} is not Sequence"
        name = inner.names[0]
        self.assertEqual(
            cst.ensure_type(
                name,
                cst.ImportAlias,
            ).name.value,
            "x",
        )

    def test_parse_import_not_an_import(self) -> None:
        with self.assertRaisesRegex(ValueError, "not an import"):
            util.parse_import("print('hello')")

    def test_parse_import_not_a_statement(self) -> None:
        with self.assertRaisesRegex(ValueError, "not a statement"):
            util.parse_import("for foo in bar: print(foo)")

    def test_parse_import_bad_syntax(self) -> None:
        with self.assertRaises(cst.ParserSyntaxError):
            util.parse_import("say 'hello'")

    def test_split_inline_comment(self) -> None:
        self.assertEqual(
            ["# foo", "# bar"], util.split_inline_comment("blah  # foo  # bar\n")
        )

    def test_split_relative(self) -> None:
        self.assertEqual(("foo", 0), util.split_relative("foo"))
        self.assertEqual(("foo", 1), util.split_relative(".foo"))
        self.assertEqual(("foo", 2), util.split_relative("..foo"))
        self.assertEqual(("foo.bar", 0), util.split_relative("foo.bar"))
        self.assertEqual(("foo.bar", 1), util.split_relative(".foo.bar"))

    def test_stem_join(self) -> None:
        self.assertEqual("foo", util.stem_join(None, "foo"))
        self.assertEqual("foo.bar", util.stem_join("foo", "bar"))
        self.assertEqual("foo.bar.baz", util.stem_join("foo", "bar.baz"))
        self.assertEqual(".bar", util.stem_join(".", "bar"))
        self.assertEqual(".foo.bar", util.stem_join(".foo", "bar"))

    def test_top_level_name(self) -> None:
        self.assertEqual("foo", util.top_level_name("foo"))
        self.assertEqual("foo", util.top_level_name("foo.bar"))
        self.assertEqual("foo", util.top_level_name("foo.bar.baz"))
        self.assertEqual("", util.top_level_name(".foo"))

    def test_with_dots(self) -> None:
        self.assertEqual("foo", util.with_dots(cst.Name(value="foo")))
        self.assertEqual(
            "foo.bar.baz",
            util.with_dots(
                cst.Attribute(
                    value=cst.Attribute(
                        value=cst.Name("foo"),
                        attr=cst.Name("bar"),
                    ),
                    attr=cst.Name("baz"),
                )
            ),
        )

        with self.assertRaisesRegex(TypeError, "Can't with_dots"):
            util.with_dots("foo.bar")  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

from usort.stdlibs import STDLIB_TOP_LEVEL_NAMES


class StdlibsTest(unittest.TestCase):
    def test_expected(self) -> None:
        # Py2 module
        self.assertNotIn("StringIO", STDLIB_TOP_LEVEL_NAMES)
        # See https://github.com/omnilib/stdlibs/issues/57
        self.assertNotIn("panel", STDLIB_TOP_LEVEL_NAMES)

        # This is specialcased, it appears to be a packaging decision for distros but I
        # don't want to include it.
        self.assertNotIn("test", STDLIB_TOP_LEVEL_NAMES)
        self.assertNotIn("lib", STDLIB_TOP_LEVEL_NAMES)
        self.assertNotIn("foo", STDLIB_TOP_LEVEL_NAMES)

        self.assertIn("io", STDLIB_TOP_LEVEL_NAMES)
        self.assertIn("os", STDLIB_TOP_LEVEL_NAMES)

        # Py2 module
        self.assertNotIn("sre", STDLIB_TOP_LEVEL_NAMES)
        for name in ("re", "sre_parse", "_sre"):
            self.assertIn(name, STDLIB_TOP_LEVEL_NAMES)
